//  ======================================================================== //
//                                                                           //
//    File      : DirContent.h                                               //
//    Purpose   :                                                            //
//    Time-stamp: <Mon Jan/13/2003 10:19 MET Coder@ReallySoft.de>            //
//                                                                           //
//    (C) April 2002 by Ralf Westram                                         //
//                                                                           //
//    Permission to use, copy, modify, distribute and sell this software     //
//    and its documentation for any purpose is hereby granted without fee,   //
//    provided that the above copyright notice appear in all copies and      //
//    that both that copyright notice and this permission notice appear      //
//    in supporting documentation.                                           //
//                                                                           //
//    Ralf Westram makes no representations about the suitability of this    //
//    software for any purpose.  It is provided "as is" without express or   //
//    implied warranty.                                                      //
//                                                                           //
//  ======================================================================== //

#ifndef DIRCONTENT_H
#define DIRCONTENT_H

#ifndef DEFS_H
#include "defs.h"
#endif

#include <dirent.h>

#include <algorithm>
#include <list>

// #ifndef DIRECTORY_H
// #include <Directory.h>
// #endif

#if defined(DEBUG)
// #define DUMP_DIR_SCAN
#endif // DEBUG

namespace playlist {

    inline char char_lower(char c) { return tolower(c); }
    inline std::string string_lower(std::string s) { transform(s.begin(), s.end(), s.begin(), char_lower); return s; }

    // -----------------------------
    //      class FilePredicate
    // -----------------------------
    class FilePredicate {
    public:
        FilePredicate() {
        }
        virtual ~FilePredicate() {}

        virtual bool operator()(const std::string& directory, const struct dirent *entry) const = 0;
    };

    // --------------------------------------------------------------------------------
    //     class hasExtension
    // --------------------------------------------------------------------------------
    class hasExtension :  public FilePredicate {
    private:
        std::string extension;

    public:
        explicit hasExtension(const std::string& ext) : extension(string_lower(ext)) {
            //         cout << "hasExtension('" << extension << "')" << endl;
        }
        virtual ~hasExtension() {}

        virtual bool operator()(const std::string& /*dir*/, const struct dirent *entry) const;
    };

    // --------------------------------------------------------------------------------
    //     class isSubDir
    // --------------------------------------------------------------------------------
    class isSubDir : public FilePredicate {
    private:
    public:
        explicit isSubDir() {
            //         cout << "isSubDir" << endl;
        }
        virtual ~isSubDir() {}

        virtual bool operator()(const std::string& dir, const struct dirent *entry) const;
    };

    // --------------------------------------------------
    //      class OrPredicate : public FilePredicate
    // --------------------------------------------------
    class OrPredicate : public FilePredicate {
    private:
        FilePredicate *pred1, *pred2;

    public:
        explicit OrPredicate(FilePredicate *pred1_,  FilePredicate *pred2_) : pred1(pred1_), pred2(pred2_) {
            //         cout << "OrPredicate" << endl;
        }
        virtual ~OrPredicate() { delete pred1; delete pred2; }

        virtual bool operator()(const std::string& dir, const struct dirent *entry) const {
            return (*pred1)(dir, entry) || (*pred2)(dir, entry);
        }
    };

    // --------------------------------------------------------------------------------

    extern bool show_dirs_read;
    extern bool force_sort;
    extern bool force_shuffle;

    // --------------------------------------------------------------------------------

    std::string catDirFile(const std::string& dir, const std::string& file);

    template <class Predicate, class Container>
    // --------------------------------------------------------------------------------
    //     void getDirContent(std::string dir, Predicate pred, Container& cont)
    // --------------------------------------------------------------------------------
    void getDirContent(std::string dir, const Predicate& pred, Container& cont) {
        if (show_dirs_read) cout << "." << flush;

        force_sort    = false;
        force_shuffle = false;

#if defined(DUMP_DIR_SCAN)
        cout << "getDirContent[1](" << dir << ")\n";
#endif // DUMP_DIR_SCAN

        DIR *d = opendir(dir.c_str());
        if (!d) throw IOError(dir, "opendir");

        while (1) {
            struct dirent *entry = readdir(d);
            if (!entry) break;

            const char *name = entry->d_name;
#if defined(DUMP_DIR_SCAN)
            cout << "name='" << name << "'\n";
#endif // DUMP_DIR_SCAN
            if (name[0]=='.' && (name[1]==0 || (name[1]=='.' && name[2]==0))) continue; // '.' und '..'

            if (pred(dir, entry)) cont.push_back(catDirFile(dir, name));

            if (strcmp(name, SORT_FILE_NAME)==0) force_sort = true;
            else if (strcmp(name, SHUFFLE_FILE_NAME)==0) force_shuffle = true;
        }

        if (force_sort && force_shuffle) {
            throw Error(std::string("'" SORT_FILE_NAME "' and '" SHUFFLE_FILE_NAME
                               "' are not allowed together in one directory (")+dir+")");
        }

        closedir(d);
    }

    template <class Predicate1, class Container1, class Predicate2, class Container2>
    // --------------------------------------------------------------------------------------------------------------------------------
    //      void getDirContent(std::string dir, const Predicate1& pred1, Container1& cont1, const Predicate2& pred2, Container2& cont2)
    // --------------------------------------------------------------------------------------------------------------------------------
    void getDirContent(std::string dir, const Predicate1& pred1, Container1& cont1, const Predicate2& pred2, Container2& cont2) {
        if (show_dirs_read) cout << "." << flush;

        force_sort    = false;
        force_shuffle = false;

#if defined(DUMP_DIR_SCAN)
        cout << "getDirContent[2](" << dir << ")\n";
#endif // DUMP_DIR_SCAN

        DIR *d = opendir(dir.c_str());
        if (!d) throw rs::err::IOError(dir, "opendir");

        while (1) {
            struct dirent *entry = readdir(d);
            if (!entry) break;

            const char *name = entry->d_name;
#if defined(DUMP_DIR_SCAN)
            cout << "name='" << name << "'\n";
#endif // DUMP_DIR_SCAN
            if (name[0]=='.' && (name[1]==0 || (name[1]=='.' && name[2]==0))) continue; // . und ..

            if (pred1(dir, entry)) cont1.push_back(catDirFile(dir, name));
            if (pred2(dir, entry)) cont2.push_back(catDirFile(dir, name));

            if (strcmp(name, SORT_FILE_NAME)         == 0) force_sort = true;
            else if (strcmp(name, SHUFFLE_FILE_NAME) == 0) force_shuffle = true;
        }

        if (force_sort && force_shuffle) {
            throw rs::err::Error(std::string("'" SORT_FILE_NAME "' and '" SHUFFLE_FILE_NAME
                                        "' are not allowed together in one directory (")+dir+")");
        }

        closedir(d);
    }

    typedef std::list<std::string> strlist;

    template <class Predicate, class Container>
    // ------------------------------------------------------------------------------------------------
    //      void getDirContentRecursive(const std::string& dir, const Predicate& pred, Container& cont)
    // ------------------------------------------------------------------------------------------------
    void getDirContentRecursive(const std::string& dir, const Predicate& pred, Container& cont) {
        //     if (seenDirectories.find(dir) != seenDirectories.end()) {
        //         cout << "Directory '" << dir << "' was already scanned. Second scan ignored to avoid recursion.\n";
        //         return;
        //     }
        //     seenDirectories.insert(dir);

        Container subdirs;
        getDirContent(dir, pred, cont, isSubDir(), subdirs);

        bool my_force_sort    = force_sort;
        bool my_force_shuffle = force_shuffle;

        for (strlist::iterator sub = subdirs.begin(); sub != subdirs.end(); ++sub) {
            getDirContentRecursive(correctDir(*sub), pred, cont);
            my_force_sort    = my_force_sort || force_sort;
            my_force_shuffle = my_force_shuffle || force_shuffle;
        }

        force_sort    = my_force_sort;
        force_shuffle = my_force_shuffle;
    }

}; // end of namespace rs

#else
#error DirContent.h included twice
#endif // DIRCONTENT_H

