//  ======================================================================== //
//                                                                           //
//    File      : Song.cpp                                                   //
//    Purpose   :                                                            //
//    Time-stamp: <Fri Jan/10/2003 22:26 MET Coder@ReallySoft.de>            //
//                                                                           //
//    (C) 2000 by Ralf Westram                                               //
//                                                                           //
//    Permission to use, copy, modify, distribute and sell this software     //
//    and its documentation for any purpose is hereby granted without fee,   //
//    provided that the above copyright notice appear in all copies and      //
//    that both that copyright notice and this permission notice appear      //
//    in supporting documentation.                                           //
//                                                                           //
//    Ralf Westram makes no representations about the suitability of this    //
//    software for any purpose.  It is provided "as is" without express or   //
//    implied warranty.                                                      //
//                                                                           //
//  ======================================================================== //

#include <sys/unistd.h>

#include <Tools.h>
#include <Directory.h>

#include "Song.h"

#if defined(DEBUG)
// #define DUMP_INFO
#endif // DEBUG

using namespace std;
using namespace rs;
using namespace rs::err;
using namespace rs::str;
using namespace rs::file;
using namespace playlist;

static unsigned mpeg_version_table[4] = { 25, 0, 2, 1 };
static unsigned layer_table[4]        = { 0, 3, 2, 1 };

static unsigned frequency_table[3/*mpeg_version-1*/][3/*frequency_type*/] = {
    {0, 0, 0},                  // MPEG V0 (doesn't exist)
    {44100, 48000, 32000},      // MPEG V1.
    {22050, 24000, 16000}       // MPEG V2
};

static int bitrate_table[5/*bitrate_table_index*/][16/*bitrate_type*/] = {
    {0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, -1 }, // MPEG V1 layer 1
    {0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384, -1 }, // MPEG V1 layer 2
    {0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, -1 }, // MPEG V1 layer 3
    {0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256, -1 }, // MPEG V2 layer 1
    {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, -1 } // MPEG V2 layer 2+3
};

static int bitrate_table_index[3/*mpeg_version-1*/][4/*mpeg_layer*/] = {
    { -1, -1, -1, -1},          // MPEG V0 (doesn't exist)
    { -1,  0,  1,  2 },         // MPEG V1
    { -1,  3,  4,  4 }          // MPEG V2
};

static unsigned layer_value[4] = { 0, 12, 144, 144 };

static const char *stereo_mode_name[4] = {"stereo", "joint stereo", "dual channel", "single channel" };
static const char *emphasis_name[4]    = {"none", "50/15", "reserved", "CCITT J 17" };

const char *MPEG_Info::get_stereo_mode_name() const { assert(stereo_mode<4); return stereo_mode_name[stereo_mode]; }
const char *MPEG_Info::get_emphasis_name() const { assert(emphasis<4); return emphasis_name[emphasis]; }

//  --------------------------------------------------------------
//      inline size_t decode_normal_integer(unsigned char *s)
//  --------------------------------------------------------------
inline size_t decode_normal_integer(unsigned char *s) {
    size_t i = 0;
    for (int p = 0; p<4; ++p) {
        unsigned u      = s[p];
        i = (i<<8)|u;
    }
    return i;
}
//  ----------------------------------------------------------------
//      inline size_t decode_syncsafe_integer(unsigned char *s)
//  ----------------------------------------------------------------
inline size_t decode_syncsafe_integer(unsigned char *s) {
    size_t i = 0;
    for (int p = 0; p<4; ++p) {
        unsigned u      = s[p];
        assert((u&128) == 0);   // MSB has to be zero

        i = (i<<7)|(u&127);
    }
    return i;
}

// -----------------------------------
//      static void crop(char *s)
// -----------------------------------
static void crop(char *s) {
    char *t = s;
    char *f = s;

    while (*f==' ') ++f;
    while (*f) *t++ = *f++;
    *t   = *f;
    --f;
    while (f>=s && *f==' ') --f;
    f[1] = 0;
}

// start of implementation of class ID3v2:

//  ---------------------------------
//      void ID3v2::dump() const
//  ---------------------------------
void ID3v2::dump() const {
    cout << "\nDump of ID3v2-info:\n";
    cout << "-------------------\n";
    cout << "was_found=" << int(was_found) << "\n";

    cout << "title='" << title << "'\n";
    cout << "artist='" << artist << "'\n";
    cout << "album='" << album << "'\n";
    cout << "year=" << year << "\n";
    cout << "genre=" << genre << "\n";
    cout << "track=" << track << "\n";
    cout << "track_max=" << track_max << "\n";
    cout << "comment=" << comment << "\n";

    cout << "orig_artist=" << orig_artist << "\n";
    cout << "copyright=" << copyright << "\n";
    cout << "encoder=" << encoder << "\n";
    cout << "url=" << url << "\n";

    cout << "conductor=" << conductor << "\n";
    cout << "composer=" << composer << "\n";
    cout << "orchestra=" << orchestra << "\n";
    cout << "opus=" << opus << "\n";

    cout << "\n";
}

// --------------------------------------------------------
//      ID3v2::ID3v2(istream& mpeg, Warnings& warnings)
// --------------------------------------------------------
ID3v2::ID3v2(istream& mpeg, Warnings& warnings)
    : was_found(false)
{
    title     = "";
    artist    = "";
    album     = "";
    year      = "";
    genre     = "";
    track     = "";
    track_max = "";
    comment   = "";

    orig_artist = "";
    copyright   = "";
    encoder     = "";
    url         = "";

    conductor = "";
    composer  = "";
    orchestra = "";
    opus      = "";

    error_reason = "";

    // --------------------------------------------
    //      read header and test if it is id3v2
    // --------------------------------------------

    unsigned char header[10];
    mpeg.seekg(0, ios::beg);
    streampos     start = 0;

    mpeg.read(reinterpret_cast<char*>(header), 10);
    if (mpeg.gcount()            != 10) return; // can't read header
    if (memcmp(header, "ID3", 3) != 0) return; // no id3v2 header

    int id3v2_version = header[3];
    //     int id3v2_revision = header[4];

#if defined(DUMP_INFO)
    cout << "id3v2_version=" << id3v2_version << "\n";
#endif // DUMP_INFO

    if (id3v2_version > 4) {
        error_reason = strf("can't handle id3v2 with major version %i", id3v2_version);
        return;
    }

    // bool unsynchronisation = header[5]&128;
    bool extended_header      = header[5]&64;
    // bool experimental      = header[5]&32;
    bool footer_present       = header[5]&16;

    size_t    id3v2_size = decode_syncsafe_integer(header+6);
    streampos id3v2_end  = start+(long(id3v2_size)-1);
    streampos frames_end = id3v2_end-(footer_present ? 10L : 0L);

    if (extended_header) {
        mpeg.read(reinterpret_cast<char*>(header), 6);
        size_t extended_header_size = decode_normal_integer(header);
        mpeg.seekg(mpeg.tellg()+(long(extended_header_size)-6), ios::beg); // skip extended header
#if defined(DUMP_INFO)
        cout << "skipping extended header (now at pos " << mpeg.tellg() << ")\n";
#endif // DUMP_INFO
    }

    // ----------------------
    //      read frames :
    // ----------------------

    bool error     = false;
    bool skipframe = false;

    for (streampos next_frame = mpeg.tellg(); !error && next_frame<frames_end; ) {
        skipframe = false;

        mpeg.seekg(next_frame, ios::beg);
        mpeg.read(reinterpret_cast<char*>(header), 10);
        if (mpeg.gcount() != 10) {
            error_reason = "couldn't read next tag-frame";
            error = true;
            break;
        }

        char frame_id[] = "xxxx";
        memcpy(frame_id, header, 4); // get frame id

    retry:
        int zero_count = 0;
        for (int p = 0; p<4; ++p) {
            if (!isalnum(frame_id[p])) {
                if (frame_id[p] == 0) {
                    zero_count++;
                }
                else {
                    warnings.add_warning(strf("id3v2-warning: Illegal characters in frame_id (%i)", int(frame_id[p])));
                    skipframe    = true;
                    break;
                }
            }
        }

        if (error) break;

        // -------------------------------
        //      test for last frame :
        // -------------------------------

        if (zero_count == 4) break; // found frame_id consting of 4*'ASCII0' (aka last frame)
        if (zero_count != 0) {
            if (frame_id[0] == 0) break; // this is an error (interpreted as end-frame here)

            if (strcmp(frame_id, "TT2") == 0 && zero_count == 1) { // this is a common error (interpreted as "TIT2" here)
                strcpy(frame_id, "TIT2");
                goto retry;
            }

            error_reason = strf("frame_id to short ('%s')", frame_id);
            error        = true;
        }

        if (error) break;

        // -------------------------------
        //      interpret frame header
        // -------------------------------

        size_t frame_size   = decode_normal_integer(header+4); // w/o header
        next_frame          = mpeg.tellg()+long(frame_size);
        bool   compressed   = header[8]&128;
        bool   encrypted    = header[8]&64;
        bool   is_textframe = (frame_id[0] == 'T' && strcmp(frame_id+1, "XXX") != 0);
        bool   is_comment   = strcmp(frame_id, "COMM") == 0;
        bool   is_url       = strcmp(frame_id, "WXXX") == 0;

        if (skipframe) continue;
        if (!is_textframe && !is_comment && !is_url) continue; // unhandled frame-type

        // ------------------------------------------------------------------------
        //      test for unknown text_encoding, compressed or encrypted content
        // ------------------------------------------------------------------------

        unsigned char text_encoding = mpeg.get();
        if (text_encoding != 0) {
            warnings.add_warning(strf("[skipped content] in frame '%s' (unknown text encoding %i)", frame_id, int(text_encoding)));
            skipframe = true;
        }
        if (compressed) {
            warnings.add_warning(strf("[skipped content] in compressed frame '%s'", frame_id));
            skipframe = true;
        }
        if (encrypted) {
            warnings.add_warning(strf("[skipped content] in encrypted frame '%s'", frame_id));
            skipframe = true;
        }

        if (skipframe) continue;

        // -----------------------------------------------------------------------
        //      calculate text_size and position file-pointer to start of text
        // -----------------------------------------------------------------------

        size_t text_size = frame_size-1;

        if (is_comment) {
            mpeg.get(); mpeg.get(); mpeg.get(); // skip 'Language'
        }
        if (is_comment || is_url) {
            size_t short_len = 0;
            if (text_size) {
                // skip short content description
                while (1) {
                    unsigned char ch = mpeg.get();
                    if (ch == 0) break;
                    ++short_len;
                    if (short_len>text_size) {
                        warnings.add_warning(strf("unterminated 'short content description' in '%s'", frame_id));
                        skipframe = true;
                        break;
                    }
                }
                short_len += 1; // terminating 0-byte
            }
            text_size -= (is_comment ? 3 : 0) + short_len;
        }

        if (skipframe) continue;
        if (!text_size) continue; // no content

        // ------------------------------
        //      read the content
        // ------------------------------

        char   buffer[text_size+1];
        size_t pos = 0;

        while (pos<text_size) {
            unsigned char ch = mpeg.get();
            buffer[pos]      = ch;
            if (!ch) break;
            ++pos;
        }
        buffer[pos] = 0;

        // ----------------------
        //      store content
        // ----------------------

        if (strcmp(frame_id, "TRCK") == 0) {
            char *slash         = strchr(buffer, '/');
            if (slash) slash[0] = 0;

            if (track.empty()) track                  = buffer;
            if (slash && track_max.empty()) track_max = slash+1;
        }
        else {
            string *target = 0;

            if (is_comment) target  = &comment;
            else if (is_url) target = &url;

            else if (strcmp(frame_id, "TIT2") == 0) { target = &title; }
            else if (strcmp(frame_id, "TPE1") == 0) { target = &artist; }
            else if (strcmp(frame_id, "TALB") == 0) { target = &album; }
            else if (strcmp(frame_id, "TCON") == 0) { target = &genre; }
            else if (strcmp(frame_id, "TYER") == 0) { target = &year; }

            else if (strcmp(frame_id, "TOPE") == 0) { target = &orig_artist; }
            else if (strcmp(frame_id, "TCOP") == 0) { target = &copyright; }
            else if (strcmp(frame_id, "TENC") == 0) { target = &encoder; }

            else if (strcmp(frame_id, "TPE3") == 0) { target = &conductor; }
            else if (strcmp(frame_id, "TCOM") == 0) { target = &composer; }
            else if (strcmp(frame_id, "TPE2") == 0) { target = &orchestra; }
            else if (strcmp(frame_id, "TIT3") == 0) { target = &opus; }

            if (target) {
                if (target->empty()) *target = buffer;
//                 else *target                 = *target + "; " + buffer;
            }
            else {
                warnings.add_warning(strf("[skipped content] in frame '%s'", frame_id));
            }
        }
    }

    if (!error && error_reason.empty()) was_found = true;
}


// -end- of implementation of class ID3v2.

// start of implementation of class MPEG_Info:

bool MPEG_Info::prefer_id3v2                  = true;
bool MPEG_Info::expect_id3v1_track_in_comment = true;

//  -------------------------------------------------------------------------
//      MPEG_Info::MPEG_Info(const string& filename, bool scan_for_id3v2)
//  -------------------------------------------------------------------------
MPEG_Info::MPEG_Info(const string& filename, bool scan_for_id3v2)
{
    error_reading_info = true;
    has_id3v1_tag      = false;
#if defined(DUMP_INFO)
    cout << "\n--------------------\n";
    cout << "filename=" << filename << "\n";
#endif // DUMP_INFO

    ifstream mpeg(filename.c_str(), ios::in|ios::binary);
    if (!mpeg.is_open() || !mpeg.good()) {
        IOError e(filename, "ifstream");
        error_reason = e.get_message();
        return;
    }

    id3v2.SetNull();
    if (scan_for_id3v2) {
        id3v2        = new ID3v2(mpeg, warnings);
        string error = id3v2->get_error_reason();

        if (!error.empty()) {
            warnings.add_warning(strf("id3v2-error: %s", error.c_str()));
        }

        if (!id3v2->found())  {
            if (id3v2->get_error_reason() != "") {
                cerr << "While reading MP3-Info from '" << filename << "' (" << id3v2->get_error_reason() << ")\n";
            }
            id3v2.SetNull();
#if defined(DUMP_INFO)
            cout << "no id3v2-tag found\n";
#endif // DUMP_INFO
        }
    }

    // --------------------
    // read header:
    // --------------------

#if defined(DUMP_INFO)
    cout << "\n";
#endif // DUMP_INFO

    mpeg.seekg(0, ios::beg);
    unsigned char header[4];
    header[0] = 255;

    while (1) {
        unsigned char ch = mpeg.get();

        if (mpeg.eof()) {
            error_reason = strf("Can't find MP3-header in '%s'", filename.c_str());
            return;
        }
        if (!mpeg.good()) {
            IOError e(filename, "MP3-headerscan");
            error_reason = e.get_message();
            return;
        }

        if (ch == 255) {
            header[1] = mpeg.get();
            if (header[1]==250 || header[1]==251) break; // followed by 250 or 251
        }
    }

    streampos headerstart = mpeg.tellg()-2L;
    mpeg.read(reinterpret_cast<char*>(header+2), 2);

    VBR    = false;
    //typedef unsigned dword;
    frames = 0;

    assert(sizeof(frames) == 4); // assert dword is 32-bit

    // check for Xing-Header
    {
        char id[5] = "xxxx";
        mpeg.seekg(headerstart+36L, ios::beg);
        mpeg.read(id, 4);

        if (strcmp(id, "Xing") == 0) {
            VBR = true;
            mpeg.seekg(headerstart+44L, ios::beg);
            mpeg.read(reinterpret_cast<char*>(&frames), 4);

            char *bytes = (char*)&frames;
            swap(bytes[0], bytes[3]); // convert byteorder
            swap(bytes[1], bytes[2]);

#if defined(DUMP_INFO)
            cout << "Xing-Header detected!\n";
            cout << "frames=" << frames << "\n";
#endif // DUMP_INFO
        }
    }

    mpeg_version = mpeg_version_table[(header[1] & 0x18) >> 3];
    layer        = layer_table[(header[1] & 0x06) >> 1];
    stereo_mode  = (header[3] & 0xc0) >> 6;     assert(stereo_mode<4 );
    emphasis     = (header[3] & 0x03);          assert(emphasis<4);

    unsigned frequency_type = (header[2] & 0x0c) >> 2;
    unsigned bitrate_type   = (header[2] & 0xf0) >> 4; // 4bit

#if defined(DUMP_INFO)
    cout << "mpeg_version=" << mpeg_version << " layer=" << layer << "\n";
    cout << "stereo_mode=" << stereo_mode << " aka '" << stereo_mode_name[stereo_mode] << "'\n";
    cout << "emphasis=" << emphasis << " aka '" << emphasis_name[emphasis] << "'\n";
    cout << "frequency_type=" << frequency_type << " bitrate_type=" << bitrate_type << "\n";
#endif

    if (frequency_type == 3) {
        frequency_type = 1; // @@@ *hack* just assuming it's the same
    }

    assert(frequency_type<3);
    int mpeg_version_index                         = mpeg_version;
    if (mpeg_version_index==25) mpeg_version_index = 2;
    if (mpeg_version_index>2) mpeg_version_index   = 0;

#if defined(DUMP_INFO)
    cout << "mpeg_version_index=" << mpeg_version_index << "\n";
#endif // DUMP_INFO

    frequency = frequency_table[mpeg_version_index][frequency_type];

    mpeg.seekg(0, ios::end);
    streampos filesize = mpeg.tellg();

    if (VBR) {
#if defined(DUMP_INFO)
        cout << "Variable bitrate!\n";
        cout << "filesize=" << filesize << "\n";
#endif // DUMP_INFO

        bitrate = unsigned((double(filesize)/frames) * (double(frequency)/1000) / layer_value[layer] + 0.5);
        //            unsigned(double(filesize*frequency) / double(frames*1000*layer_value[layer]) + 0.5);
    }
    else {
        int idx = bitrate_table_index[mpeg_version_index][layer];
        bitrate = 1;            // e.g. for VBR
        if (idx>=0) {
            int value              = bitrate_table[idx][bitrate_type];
            if (value!=-1) bitrate = value;
        }
    }

    if (bitrate) {
        unsigned ms = (filesize*8) / bitrate;
#if defined(DUMP_INFO)
        cout << "duration=" << ms << " ms\n";
#endif // DUMP_INFO
        duration    = (ms+500)/1000;
    }
    else {
        duration = 11*60+11; // just a number (11:11)
    }

#if defined(DUMP_INFO)
    cout << "frequency=" << frequency << " Hz - bitrate=" << bitrate << " kHz - duration=" << duration << " sec.\n";
#endif // DUMP_INFO

    original   = (header[3] & 0x04) >> 2;
    copyright  = (header[3] & 0x08) >> 3;
    padding    = (header[2] & 0x02) >> 1;
    is_Private = (header[2] & 0x01);
    CRC        = !(header[1] & 0x01);

#if defined(DUMP_INFO)
    cout << "original=" << original << " copyright=" << copyright << " padding=" << padding
         << " is_private=" << is_Private << " CRC=" << CRC << "\n";
#endif // DUMP_INFO

    // --------------------
    // read tags:
    // --------------------

    mpeg.seekg(-128, ios::end); // goto start of tags

    memset(&id3v1, 0, sizeof(id3v1));

    mpeg.read(id3v1.tag, TAGLEN_TAG);
    mpeg.read(id3v1.songname, TAGLEN_SONG);
    mpeg.read(id3v1.artist, TAGLEN_ARTIST);
    mpeg.read(id3v1.album, TAGLEN_ALBUM);
    mpeg.read(id3v1.year, TAGLEN_YEAR);
    mpeg.read(id3v1.comment, TAGLEN_COMMENT);
    id3v1.genre = mpeg.get();

    crop(id3v1.tag);
    crop(id3v1.songname);
    crop(id3v1.artist);
    crop(id3v1.album);
    crop(id3v1.year);
    crop(id3v1.comment);

    has_id3v1_tag = strcmp(id3v1.tag, "TAG") == 0;

#if defined(DUMP_INFO)
    if (has_id3v1()) {
        cout << "Found ID3v1:\n"; // " TAG='" << id3.tag << "'\n";
        cout << "songname='" << get_v1_songname() << "'\n";
        cout << "artist  ='" << get_v1_artist() << "'\n";
        cout << "album   ='" << get_v1_album() << "'\n";
        cout << "year    ='" << get_v1_year() << "'\n";
        cout << "track   ='" << get_v1_track() << "'\n";
        cout << "comment   ='" << get_v1_comment() << "'\n";
    }
    if (has_id3v2()) {
        cout << "Found ID3v2:\n";
        cout << "songname='" << get_v2_songname() << "'\n";
        cout << "artist  ='" << get_v2_artist() << "'\n";
        cout << "album   ='" << get_v2_album() << "'\n";
        cout << "year    ='" << get_v2_year() << "'\n";
        cout << "track   ='" << get_v2_track() << "'\n";
        cout << "comment   ='" << get_v2_comment() << "'\n";
    }
    if (!has_id3v1() && !has_id3v2()) {
        cout << "no id3v1- or id3v2-tag found\n";
    }
#endif

    check_consistency();

    error_reading_info = false;
    error_reason       = "";
}

const string& MPEG_Info::get_v1_songname() const { static string songname; songname = ""; if (has_id3v1()) songname = id3v1.songname; return songname; }
const string& MPEG_Info::get_v1_artist() const { static string   artist; artist     = ""; if (has_id3v1()) artist = id3v1.artist; return artist; }
const string& MPEG_Info::get_v1_album() const { static string    album; album       = ""; if (has_id3v1()) album = id3v1.album; return album; }
const string& MPEG_Info::get_v1_year() const { static string     year; year         = ""; if (has_id3v1()) year = id3v1.year; return year; }
const string& MPEG_Info::get_v1_comment() const { static string     comment; comment         = ""; if (has_id3v1()) comment = id3v1.comment; return comment; }


const string& MPEG_Info::get_v1_track() const {
    static string track;
    track               = "";
    if (has_id3v1()) {
        if (expect_id3v1_track_in_comment && id3v1.comment[TAGLEN_COMMENT-2] == 0) {
            int tracknr = id3v1.comment[TAGLEN_COMMENT-1];
            if (tracknr>0) track = strf("%i", tracknr);
        }
    }
    return track;
}

static string empty = "";

const string& MPEG_Info::get_v2_songname() const { return has_id3v2() ? id3v2->get_songname() : empty; }
const string& MPEG_Info::get_v2_artist() const { return has_id3v2() ? id3v2->get_artist() : empty; }
const string& MPEG_Info::get_v2_album() const { return has_id3v2() ? id3v2->get_album() : empty; }
const string& MPEG_Info::get_v2_year() const { return has_id3v2() ? id3v2->get_year() : empty; }
const string& MPEG_Info::get_v2_track() const { return has_id3v2() ? id3v2->get_track() : empty; }
const string& MPEG_Info::get_v2_track_max() const { return has_id3v2() ? id3v2->get_track_max() : empty;}

const string& MPEG_Info::get_v2_comment() const { return has_id3v2() ? id3v2->get_comment() : empty; }

//  ---------------------------------------------------------------------------------------------------------------------------
//      inline const string& MPEG_Info::chooseContentToUse(const string& id3v1_Content, const string& id3v2_Content) const
//  ---------------------------------------------------------------------------------------------------------------------------
inline const string& MPEG_Info::chooseContentToUse(const string& id3v1_Content, const string& id3v2_Content) const {
    if (prefer_id3v2) {
        if (has_id3v2()) return id3v2_Content;
    }
    return id3v1_Content;
}
const string& MPEG_Info::get_songname() const { return chooseContentToUse(get_v1_songname(), get_v2_songname()); }
const string& MPEG_Info::get_artist() const { return chooseContentToUse(get_v1_artist(), get_v2_artist()); }
const string& MPEG_Info::get_album() const { return chooseContentToUse(get_v1_album(), get_v2_album()); }
const string& MPEG_Info::get_year() const { return chooseContentToUse(get_v1_year(), get_v2_year()); }
const string& MPEG_Info::get_track() const { return chooseContentToUse(get_v1_track(), get_v2_track()); }
const string& MPEG_Info::get_comment() const { return chooseContentToUse(get_v1_comment(), get_v2_comment()); }

// the following entries do not exist in id3v1 :
const string& MPEG_Info::get_orig_artist() const { return has_id3v2() ? id3v2->get_orig_artist() : empty; }
const string& MPEG_Info::get_copyright() const { return has_id3v2() ? id3v2->get_copyright() : empty; }
const string& MPEG_Info::get_encoder() const { return has_id3v2() ? id3v2->get_encoder() : empty; }
const string& MPEG_Info::get_url() const { return has_id3v2() ? id3v2->get_url() : empty; }

const string& MPEG_Info::get_conductor() const { return has_id3v2() ? id3v2->get_conductor() : empty; }
const string& MPEG_Info::get_composer() const { return has_id3v2() ? id3v2->get_composer() : empty; }
const string& MPEG_Info::get_orchestra() const { return has_id3v2() ? id3v2->get_orchestra() : empty; }
const string& MPEG_Info::get_opus() const { return has_id3v2() ? id3v2->get_opus() : empty; }

static const char *genre_name[] = {
    "Blues", // 0
    "Classic Rock", // 1
    "Country", // 2
    "Dance", // 3
    "Disco", // 4
    "Funk", // 5
    "Grunge", // 6
    "Hip-Hop", // 7
    "Jazz", // 8
    "Metal", // 9
    "New Age", // 10
    "Oldies", // 11
    "Other", // 12
    "Pop", // 13
    "R&B", // 14
    "Rap", // 15
    "Reggae", // 16
    "Rock", // 17
    "Techno", // 18
    "Industrial", // 19
    "Alternative", // 20
    "Ska", // 21
    "Death Metal", // 22
    "Pranks", // 23
    "Soundtrack", // 24
    "Euro-Techno", // 25
    "Ambient", // 26
    "Trip-Hop", // 27
    "Vocal", // 28
    "Jazz+Funk", // 29
    "Fusion", // 30
    "Trance", // 31
    "Classical", // 32
    "Instrumental", // 33
    "Acid", // 34
    "House", // 35
    "Game", // 36
    "Sound Clip", // 37
    "Gospel", // 38
    "Noise", // 39
    "AlternRock", // 40
    "Bass", // 41
    "Soul", // 42
    "Punk", // 43
    "Space", // 44
    "Meditative", // 45
    "Instrumental Pop", // 46
    "Instrumental Rock", // 47
    "Ethnic", // 48
    "Gothic", // 49
    "Darkwave", // 50
    "Techno-Industrial", // 51
    "Electronic", // 52
    "Pop-Folk", // 53
    "Eurodance", // 54
    "Dream", // 55
    "Southern Rock", // 56
    "Comedy", // 57
    "Cult", // 58
    "Gangsta", // 59
    "Top 40", // 60
    "Christian Rap", // 61
    "Pop/Funk", // 62
    "Jungle", // 63
    "Native American", // 64
    "Cabaret", // 65
    "New Wave", // 66
    "Psychedelic", // 67
    "Rave", // 68
    "Showtunes", // 69
    "Trailer", // 70
    "Lo-Fi", // 71
    "Tribal", // 72
    "Acid Punk", // 73
    "Acid Jazz", // 74
    "Polka", // 75
    "Retro", // 76
    "Musical", // 77
    "Rock & Roll", // 78
    "Hard Rock", // 79
    "Folk", // 80
    "Folk-Rock", // 81
    "National Folk", // 82
    "Swing", // 83
    "Fast Fusion", // 84
    "Bebob", // 85
    "Latin", // 86
    "Revival", // 87
    "Celtic", // 88
    "Bluegrass", // 89
    "Avantgarde", // 90
    "Gothic Rock", // 91
    "Progressive Rock", // 92
    "Psychedelic Rock", // 93
    "Symphonic Rock", // 94
    "Slow Rock", // 95
    "Big Band", // 96
    "Chorus", // 97
    "Easy Listening", // 98
    "Acoustic", // 99
    "Humour", // 100
    "Speech", // 101
    "Chanson", // 102
    "Opera", // 103
    "Chamber Music", // 104
    "Sonata", // 105
    "Symphony", // 106
    "Booty Bass", // 107
    "Primus", // 108
    "Porn Groove", // 109
    "Satire", // 110
    "Slow Jam", // 111
    "Club", // 112
    "Tango", // 113
    "Samba", // 114
    "Folklore", // 115
    "Ballad", // 116
    "Power Ballad", // 117
    "Rhythmic Soul", // 118
    "Freestyle", // 119
    "Duet", // 120
    "Punk Rock", // 121
    "Drum Solo", // 122
    "A capella", // 123
    "Euro-House", // 124
    "Dance Hall", // 125
};

#define GENRE_ID_MIN 0
#define GENRE_ID_MAX 125

//  -------------------------------------------------
//      static const char *genreId2name(int id)
//  -------------------------------------------------
static const char *genreId2name(int id)  {
    assert(genre_name[125] == "Dance Hall");
    if (id <= GENRE_ID_MIN || id >= GENRE_ID_MAX) return 0;
// #if defined(DEBUG)
//     static char *genre_debug_name;
//     if (genre_debug_name) delete genre_debug_name;
//     genre_debug_name = new char[strlen(genre_name[id])+5];
//     sprintf(genre_debug_name, "[1] %s", genre_name[id]);
//     return genre_debug_name;
// #else
     return genre_name[id];
// #endif // DEBUG
}

//  ------------------------------------------------------
//      const string& MPEG_Info::get_v1_genre() const
//  ------------------------------------------------------
const string& MPEG_Info::get_v1_genre() const {
    static string genre;
    genre = "";
    if (has_id3v1()) {
        const char *g = genreId2name(id3v1.genre);
        if (g) {
            genre = g;
        }
    }
    return genre;
}
//  ---------------------------------------------------------
//      const genreSet& MPEG_Info::get_v2_genres() const
//  ---------------------------------------------------------
const genreSet& MPEG_Info::get_v2_genres() const {
    static genreSet genres;
    genres.clear();

    if (has_id3v2()) {
        static string genre_string;
        size_t        scanPos = 0;
        genre_string          = id3v2->get_genre();

        while (1) {
            size_t open_paren = genre_string.find('(', scanPos);
            if (open_paren == string::npos) { // no '(' in string
                string name = genre_string.substr(scanPos);
                if (name != "") genres.insert(name);
                break;
            }

            if (open_paren>scanPos) { // text in front of '('
                string name = genre_string.substr(scanPos, open_paren-scanPos);
                if (name != "") genres.insert(name);
                scanPos     = open_paren;
            }
            assert(open_paren == scanPos);

            size_t close_paren = genre_string.find(')', open_paren+1);
            if (close_paren == string::npos) break; // no ')' -> abort

            string paren_content = genre_string.substr(open_paren+1, close_paren-open_paren-1);
            int    id            = atoi(paren_content.c_str());
            if (id != 0) {
                const char *name = genreId2name(id);
                if (name) genres.insert(name);
            }

            scanPos = close_paren+1;
        }
    }
    return genres;
}
//  ------------------------------------------------------
//      const genreSet& MPEG_Info::get_genres() const
//  ------------------------------------------------------
const genreSet& MPEG_Info::get_genres() const {
    static genreSet genres;
    genres = get_v2_genres();

    const string v1 = get_v1_genre();
    if (v1 != "") genres.insert(v1);

    return genres;
}

// ------------------------------------------------------------------------------------------------------------------------
//      static bool accept_v1_as_prefix_of_v2(const char *tagname, const string& content_v1, const string& content_v2)
// ------------------------------------------------------------------------------------------------------------------------
static bool accept_v1_as_prefix_of_v2(const char */*tagname*/, const string& content_v1, const string& content_v2) {
    return content_v2.find(content_v1) == 0;
}

// ----------------------------------------------------------------------------------------------------------------
//      static bool accept_special_tracks(const char *tagname, const string& track_v1, const string& track_v2)
// ----------------------------------------------------------------------------------------------------------------
static bool accept_special_tracks(const char */*tagname*/, const string& track_v1, const string& track_v2) {
    int val_v1 = atoi(track_v1.c_str());
    int val_v2 = atoi(track_v2.c_str());

    return val_v1 == val_v2;
}

// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//      void MPEG_Info::check_content_match(const char *tagname, const string& content_v1, const string& content_v2, bool (*accept)(const char *, const string&, const string&))
// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
void MPEG_Info::check_content_match(const char *tagname, const string& content_v1, const string& content_v2, bool (*accept)(const char *, const string&, const string&))
{
    if (content_v1 == content_v2) return;
    if (!content_v2.empty()) {
        if (content_v1.empty()) return;
        if (accept(tagname, content_v1, content_v2)) return;

        cout << tagname << "-mismatch (" << content_v1 << "/" << content_v2 << ")\n" << flush;

        string warn_mess = strf("mismatch between v1 and v2 [%s]", tagname);
        warnings.add_warning(warn_mess);
    }
}

// --------------------------------------------
//      void MPEG_Info::check_consistency()
// --------------------------------------------
void MPEG_Info::check_consistency()
{
    if (has_id3v2()) {
        // check if tags of v1 and v2 contain the same

        check_content_match("songname", get_v1_songname(), get_v2_songname(), accept_v1_as_prefix_of_v2);
        check_content_match("artist", get_v1_artist(), get_v2_artist(), accept_v1_as_prefix_of_v2);
        check_content_match("album", get_v1_album(), get_v2_album(), accept_v1_as_prefix_of_v2);
        check_content_match("comment", get_v1_comment(), get_v2_comment(), accept_v1_as_prefix_of_v2);

        check_content_match("track", get_v1_track(), get_v2_track(), accept_special_tracks);

        check_content_match("year", get_v1_year(), get_v2_year(), 0);

        // @@@ genre check ?

        // check if track <= track_max

        const string& track_str     = get_v2_track();
        const string& track_max_str = get_v2_track_max();
        int           track         = atoi(track_str.c_str());
        int           track_max     = atoi(track_max_str.c_str());
        bool          compare       = true;

        if (track < 1) {
            if (!track_str.empty()) warnings.add_warning("illegal tracknumber");
            compare = false;
        }
        if (track_max < 1) {
            if (!track_max_str.empty()) warnings.add_warning("illegal maximum tracknumber");
            compare = false;
        }

        if (compare && (track > track_max)) {
            warnings.add_warning("tracknumber above maximum");
        }
    }
    else if (has_id3v1()) {
        warnings.add_warning("only v1 tags");
    }
    else {
        warnings.add_warning("no tags at all");
    }
}


// -end- of implementation of class MPEG_Info.

// start of implementation of class Warnings:

Warnings::WarningMap Warnings::warnings;

// --------------------------------------------------------------
//      const string& Warnings::get_warning(size_t idx) const
// --------------------------------------------------------------
const string& Warnings::get_warning(size_t idx) const
{
    assert(idx<size());

    NumberSet::const_iterator found = numbers.begin();
    advance(found, int(idx));
    assert(found != numbers.end());
    if (found != numbers.end()) {
        for (WarningMap::const_iterator wmi = warnings.begin(); wmi != warnings.end(); ++wmi) {
            if (wmi->second == *found) return wmi->first;
        }
    }
    assert(0);          // internal error (asked for non-existing index)
    static string err = "<internal error>";
    return err;
}

// ----------------------------------------------------------
//      void Warnings::add_warning(const string& message)
// ----------------------------------------------------------
void Warnings::add_warning(const string& message)
{
    WarningMap::const_iterator found = warnings.find(message);
    int                        idx;
    if (found == warnings.end()) { // a new warning message
        idx               = warnings.size();
        warnings[message] = idx;
    }
    else {
        idx = found->second;
    }
    numbers.insert(idx);
}

// -end- of implementation of class Warnings.

// start of implementation of class Song:

// ------------------------------
//      void Song::testLink()
// ------------------------------
void Song::testLink()
{
    isWindowsLink = rs::file::isLink(filename_maybe_link, linkTarget);
    if (isWindowsLink) {
        struct stat st;
        int         result = ::stat(linkTarget.c_str(), &st);
        isBrokenLink       = (result != 0);
    }
    else {
        isBrokenLink = false;
    }

#if defined(DEBUG) && 0
    cout << "filename_maybe_link='" << filename_maybe_link << "'\n";
    cout << "linkTarget='" << linkTarget << "'\n";
    cout << "isWindowsLink=" << isWindowsLink << "\n";
    cout << "isBrokenLink=" << isBrokenLink << "\n";
#endif // DEBUG
}

// -------------------------------------------------------------
//      void Song::scan(bool read_info, bool scan_for_id3v2)
// -------------------------------------------------------------
void Song::scan(bool read_info, bool scan_for_id3v2)
{
    assert(!isBrokenLink);
    filedate = rs::posix::getFileModTime(get_filename());
    if (read_info) {
        mpeg = new MPEG_Info(get_filename(), scan_for_id3v2);
#if defined(DEBUG)
        if (mpeg->infoError()) {
            cerr << "\nError scanning MP3-Info (reason: " << mpeg->getErrorReason() << ")";
        }
#endif // DEBUG
    }
}

// -end- of implementation of class Song.

