/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// Copyright (C) 2000                                                      //
// Ralf Westram                                                            //
// (Coded@ReallySoft.de)                                                   //
//                                                                         //
// Permission to use, copy, modify, distribute and sell this software      //
// and its documentation for any purpose is hereby granted without fee,    //
// provided that the above copyright notice appear in all copies and       //
// that both that copyright notice and this permission notice appear       //
// in supporting documentation.  Ralf Westram makes no                     //
// representations about the suitability of this software for any          //
// purpose.  It is provided "as is" without express or implied warranty.   //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

#ifndef SONG_H
#define SONG_H

#ifndef __LIST__
#include <list>
#endif
#ifndef __SET__
#include <set>
#endif
#ifndef MAP
#include <map>
#endif

#ifndef SMARTPTR_H
#include <smartptr.h>
#endif
#ifndef SYS_DEP_H
#include <sys_dep.h>
#endif

#ifndef DEFS_H
#include "defs.h"
#endif



namespace playlist {

#define TAGLEN_TAG     3
#define TAGLEN_SONG    30
#define TAGLEN_ARTIST  30
#define TAGLEN_ALBUM   30
#define TAGLEN_YEAR    4
#define TAGLEN_COMMENT 30
#define TAGLEN_GENRE   1


    //  ---------------------------------
    //      typedef struct ID3_struct
    //  ---------------------------------
    typedef struct ID3_struct
    {
        char tag[TAGLEN_TAG+1];
        char songname[TAGLEN_SONG+1];
        char artist[TAGLEN_ARTIST+1];
        char album[TAGLEN_ALBUM+1];
        char year[TAGLEN_YEAR+1];
        char comment[TAGLEN_COMMENT+1];
        int  genre;
    };

    // ------------------------
    //      class Warnings
    // ------------------------
    class Warnings {
    private:
        typedef std::set<size_t>         NumberSet;
        typedef std::map<std::string, size_t> WarningMap;

        NumberSet         numbers;
        static WarningMap warnings;

    public:
        Warnings() {}
        virtual ~Warnings() {}

        void add_warning(const std::string& message);
        size_t size() const { return numbers.size(); }
        const std::string& get_warning(size_t idx) const;
    };

    //  --------------------
    //      class id3v2
    //  --------------------
    class ID3v2 {
    private:
        bool was_found;

        std::string title;
        std::string artist;
        std::string album;
        std::string year;
        std::string genre;
        std::string track;
        std::string track_max;       // maximum number of tracks
        std::string comment;

        std::string orig_artist;
        std::string copyright;
        std::string encoder;
        std::string url;

        std::string conductor;
        std::string composer;
        std::string orchestra;
        std::string opus;

        std::string error_reason;

    public:
        ID3v2(std::istream& mpeg, Warnings& warnings);
        virtual ~ID3v2() {}

        bool found() const { return was_found; }
        void dump() const;
        const std::string& get_error_reason() const { return error_reason; }

        const std::string& get_songname() const { return title; }
        const std::string& get_artist() const { return artist; }
        const std::string& get_album() const { return album; }
        const std::string& get_year() const { return year; }
        const std::string& get_genre() const { return genre; }
        const std::string& get_track() const { return track; }
        const std::string& get_track_max() const { return track_max; }
        const std::string& get_comment() const { return comment; }

        const std::string& get_orig_artist() const { return orig_artist; }
        const std::string& get_copyright() const { return copyright; }
        const std::string& get_encoder() const { return encoder; }
        const std::string& get_url() const { return url; }

        const std::string& get_conductor() const { return conductor; }
        const std::string& get_composer() const { return composer; }
        const std::string& get_orchestra() const { return orchestra; }
        const std::string& get_opus() const { return opus; }
    };

    typedef rs::SmartPtr<ID3v2> ID3v2Ptr;
    typedef std::set<std::string>     genreSet;

    //  ------------------------
    //      class MPEG_Info
    //  ------------------------
    class MPEG_Info {
    private:

        bool   error_reading_info; // if this is true only 'error_reason' is valid
        std::string error_reason;

        unsigned short mpeg_version;
        unsigned short layer;

        unsigned short stereo_mode;
        unsigned short emphasis;

        bool     VBR;
        unsigned frames;            // only valid if VBR == true

        unsigned frequency;
        unsigned bitrate;
        unsigned duration;

        bool original;
        bool copyright;
        bool is_Private;

        bool padding;
        bool CRC;

        ID3_struct id3v1;
        bool       has_id3v1_tag;
        ID3v2Ptr   id3v2;       // if id3v2-info was found

        inline const std::string& chooseContentToUse(const std::string& id3v1_Content, const std::string& id3v2_Content) const;

        Warnings warnings;

        void check_content_match(const char *tagname, const std::string& content_v1, const std::string& content_v2, bool (*accept)(const char *, const std::string&, const std::string&));
        void check_consistency();

    public:

        MPEG_Info(const std::string& filename, bool scan_for_id3v2);
        virtual ~MPEG_Info() {}

        bool infoError() const { return error_reading_info; }
        const std::string& getErrorReason() const { assert(infoError()); return error_reason; }

        size_t no_of_warnings() const { return warnings.size(); }
        const std::string& get_warning(size_t idx) const { return warnings.get_warning(idx); }

        unsigned get_mpeg_version() const { return mpeg_version; }
        unsigned get_mpeg_layer() const { return layer; }

        unsigned get_stereo_mode() const { return stereo_mode; }
        const char *get_stereo_mode_name() const;

        unsigned get_emphasis() const { return emphasis; }
        const char *get_emphasis_name() const;

        bool is_VBR() const { return VBR; }

        unsigned get_frequency() const { return frequency; } // hz
        unsigned get_bitrate() const { return bitrate; }    // khz
        unsigned get_duration() const { return duration; } // seconds

        bool is_original() const { return original; }
        bool is_copyrighted() const { return copyright; }
        bool is_private() const { return is_Private; }

        bool has_padding() const { return padding; }
        bool has_CRC() const { return CRC; }

        // ID3

        bool has_id3v1() const { return has_id3v1_tag; }
        bool has_id3v2() const { return !id3v2.Null(); }
        bool has_id3() const { return has_id3v1() || has_id3v2(); }

        static bool prefer_id3v2; // true
        static bool expect_id3v1_track_in_comment; // true

        // following functions return "" if not tag set
        const std::string& get_v1_songname() const;
        const std::string& get_v1_artist() const;
        const std::string& get_v1_album() const;
        const std::string& get_v1_year() const;
        const std::string& get_v1_track() const;
        const std::string& get_v1_genre() const;
        const std::string& get_v1_comment() const;

        const std::string& get_v2_songname() const;
        const std::string& get_v2_artist() const;
        const std::string& get_v2_album() const;
        const std::string& get_v2_year() const;
        const std::string& get_v2_track() const;
        const std::string& get_v2_track_max() const;
        const std::string& get_v2_comment() const;

        const genreSet& get_v2_genres() const;

        // result of the following functions depends on 'prefer_id3v2'
        const std::string& get_songname() const;
        const std::string& get_artist() const;
        const std::string& get_album() const;
        const std::string& get_year() const;
        const std::string& get_track() const;
        const std::string& get_comment() const;

        // id3v2 only tags :
        const std::string& get_orig_artist() const;
        const std::string& get_copyright() const;
        const std::string& get_encoder() const;
        const std::string& get_url() const;

        const std::string& get_conductor() const;
        const std::string& get_composer() const;
        const std::string& get_orchestra() const;
        const std::string& get_opus() const;

        const genreSet& get_genres() const;
    };

    typedef rs::SmartPtr<MPEG_Info> MPEG_InfoPtr;

    //  -------------------
    //      class Song
    //  -------------------
    class Song {
    private:

        std::string       filename_maybe_link;
        time_t       filedate;  // unix-date of MP3-file
        MPEG_InfoPtr mpeg;
//         mutable bool is_on_Album;

        bool   isWindowsLink;
        bool   isBrokenLink;
        std::string linkTarget;

        Song(const Song&) { assert(0); }

        void testLink();

        void scan(bool read_info, bool scan_for_id3v2);

    public:
        const std::string& get_filename() const { assert(!isBrokenLink); return isWindowsLink ? linkTarget : filename_maybe_link; }
        const std::string& get_filename_maybe_link() const { return filename_maybe_link; }
        const std::string& get_link_target() const { return linkTarget; }

        time_t get_filedate() const { assert(!isBrokenLink); return filedate; }

        bool isBroken() const { return isWindowsLink && isBrokenLink; }
        bool isLink() const { return isWindowsLink; }

        Song(const std::string& fname, bool read_info, bool scan_for_id3v2)
            : filename_maybe_link(fname)
//             , is_on_Album(false)
        {
            if (rs::show_debug_info) std::cout << "Begin song '" << fname << "'\n";
            testLink();
            if (!isBrokenLink) scan(read_info, scan_for_id3v2);
            if (rs::show_debug_info) std::cout << "End song '" << fname << "'\n";
        }
        virtual ~Song() {}

//         bool IsOnAlbum() const { return is_on_Album; }
//         void set_IsOnAlbum(bool is_on_Album_) const { is_on_Album = is_on_Album_; }

        bool operator<(const Song& other) const { return filename_maybe_link<other.filename_maybe_link; }

        bool sameTarget(const Song& other) const { return get_filename() == other.get_filename(); }

        bool has_MPEG_info() const { return !mpeg.Null(); }

        bool          infoError() const { return has_MPEG_info() && mpeg->infoError(); }
        const std::string& getErrorReason() const { assert(infoError()); return mpeg->getErrorReason(); }
        void removeCorruptMPEGinfo()  { assert(infoError()); mpeg.SetNull(); }

        const MPEG_Info& get_MPEG_Info() const { return *mpeg; }
    };

    typedef rs::SmartPtr<Song> SongPtr;
    typedef std::list<SongPtr> SongList;

};

#else
#error Song.h included twice
#endif // SONG_H

