//  ======================================================================== //
//                                                                           //
//    File      : SongContainer.cpp                                          //
//    Purpose   :                                                            //
//    Time-stamp: <Sun Feb/02/2003 09:59 MET Coder@ReallySoft.de>            //
//                                                                           //
//    (C) April 2002 by Ralf Westram                                         //
//                                                                           //
//    Permission to use, copy, modify, distribute and sell this software     //
//    and its documentation for any purpose is hereby granted without fee,   //
//    provided that the above copyright notice appear in all copies and      //
//    that both that copyright notice and this permission notice appear      //
//    in supporting documentation.                                           //
//                                                                           //
//    Ralf Westram makes no representations about the suitability of this    //
//    software for any purpose.  It is provided "as is" without express or   //
//    implied warranty.                                                      //
//                                                                           //
//  ======================================================================== //

#include <Directory.h>

#include "SongContainer.h"
#include "DirContent.h"
#include "save.h"
#include <Tools.h>

using namespace std;
using namespace rs;
using namespace rs::err;
using namespace rs::str;
using namespace playlist;

// -------------------------
//      class Tokenizer
// -------------------------
class Tokenizer {
private:
    const string& self;
    const string  separator;
    size_t        start;        // always points to the start of the next element (if == string::npos -> got all tokens)
    size_t        found;        // always points to the first separator behind start
    bool          empty_tokens_allowed;

    bool gotAll() const { return start == string::npos; }

    void next() {
        if (!gotAll()) {
            if (found == string::npos) {
                start = found;
            }
            else {
                if (found) start = found+1;
                found            = self.find_first_of(separator, start);
            }
        }
    }

public:
    Tokenizer(const string& s, const string& separator_)
        : self(s)
        , separator(separator_)
        , start(0)
        , found(0)
        , empty_tokens_allowed(false)
    {
        next();
    }
    virtual ~Tokenizer() {}

    void allow_empty_tokens() { empty_tokens_allowed = true; }

    string get() {
        string s;
        if (!gotAll()) s = self.substr(start, found-start);
        next();
        if (s.empty() && !empty_tokens_allowed) throw Error(strf("Empty token in '%s'", s.c_str()));
        return s;
    }

    operator bool () const { return !gotAll(); }
};

// --------------------------------------------------------------------------------

static FilePredicate *pred_songs           = 0;
static FilePredicate *pred_playlists       = 0;
static FilePredicate *pred_songs_and_links = 0;
static FilePredicate *pred_subdir          = 0;

namespace playlist {
    void initPredicates(const string& extension_list) {
        Tokenizer tok(extension_list, ";");
        assert(tok);

        while (tok) {
            string token               = tok.get();
            if (!token.length()) throw Error("Empty token in extension list");
            if (token[0] != '.') token = string(".")+token;

            FilePredicate *neu = new hasExtension(token);
            FilePredicate *neu2 = new hasExtension(token);

            if (pred_songs) pred_songs = new OrPredicate(pred_songs, neu);
            else pred_songs            = neu;

            token += ".lnk";
            neu    = new OrPredicate(new hasExtension(token), neu2);

            if (pred_songs_and_links) pred_songs_and_links = new OrPredicate(pred_songs_and_links, neu);
            else pred_songs_and_links                      = neu;
        }

        assert(pred_songs);
        assert(pred_songs_and_links);

        pred_subdir    = new isSubDir();
        pred_playlists = new hasExtension(PLAYLIST_EXTENSION);
    };
}

// --------------------------------------------------------------------------------

// start of implementation of class SongContainer:

string SongContainer::all_name;
string SongContainer::rest_name;
string SongContainer::rest_postfix;
bool   SongContainer::top_down_naming = false;

void SongContainer::init(const string& allrest_def, bool top_down_naming_) {
    Tokenizer tok(allrest_def, ";");
    string miss_after;

    if (tok) {
        all_name = tok.get();
        if (tok) {
            rest_name = tok.get();
            if (tok) {
                rest_postfix = tok.get();
                if (tok) throw Error(strf("Unexpected ';' after '%s'", rest_postfix.c_str()));
            }
            else miss_after = rest_name;
        }
        else miss_after = all_name;
    }
    else {
        throw Error("Definition missing");
    }

    if (!miss_after.empty()) throw Error(strf("Missing ';' after '%s'", miss_after.c_str()));

    top_down_naming = top_down_naming_;
}

// ----------------------------------------------------------
//      void SongContainer::merge(SongContainerPtr other)
// ----------------------------------------------------------
void SongContainer::merge(SongContainerPtr other)
{
//     songs.merge(other->songs);

    COPY_ALL(other->songs, songs);

    force_sorted   = force_sorted || other->force_sorted;
    force_shuffled = force_shuffled || other->force_shuffled;

    for (SongContainerList::iterator s = other->son_begin(); s != other->son_end(); ++s) { // merge all sons of other
        SongContainerList::iterator found = sons.end();
        for (SongContainerList::iterator s2 = son_begin(); s2 != son_end(); ++s2) { // search for son with same name
            if ((*s2)->get_name() == (*s)->get_name()) {
                found = s2;
                break;
            }
        }
        if (found != sons.end()) { // merge two sons
            (*found)->merge(*s);
        }
        else {                  // insert new son
            (*s)->set_father(this);
            sons.push_back(*s);
        }
    }
}
// ---------------------------------------------------------------------------
//      static void renamePlaylists(const string& alt, const string& neu)
// ---------------------------------------------------------------------------
static void renamePlaylists(string alt, string neu) {
    alt += PLAYLIST_EXTENSION;
    neu += PLAYLIST_EXTENSION;

    if (rename(alt.c_str(), neu.c_str()) != 0) {
        throw IOError(alt, strf("rename to '%s'", neu.c_str()));
    }
}

// -----------------------------------------------------------------------------------------------
//      bool SongContainer::savePlaylists(SongList& all_songs, size_t min_playlist_size) const
// -----------------------------------------------------------------------------------------------
bool SongContainer::savePlaylists(SongList& all, size_t min_playlist_size) const
{
    SongList rest;
    COPY_ALL(songs, rest);      // put all songs not contained in subdirs to rest-list
    COPY_ALL(songs, all);       // put all songs not contained in subdirs to all-list

//     static string last_saved_name; // contains name of last saved playlist

    int subs_saved = 0;
    for (SongContainerList::const_iterator sc = son_begin(); sc != son_end(); ++sc) {
        SongList sub_songs;
        bool     saved = (*sc)->savePlaylists(sub_songs, min_playlist_size);

        COPY_ALL(sub_songs, all);

        if (saved) ++subs_saved;
        else COPY_ALL(sub_songs, rest);
    }

    string my_name = get_recursive_name();

    if (rest.size() != all.size()) {
        if (rest.size() >= min_playlist_size) {
            string restName = my_name.length() ? my_name+rest_postfix : rest_name;  // " (rest)" : "-Rest";
            if (savePlaylist(restName, rest, force_sorted, force_shuffled, min_playlist_size)) {
                ++subs_saved;
//                 last_saved_name = restName;
            }
        }
    }

    bool saved = false;
    if (all.size() >= min_playlist_size) {
        string allName = my_name.length() ? my_name : all_name; // "-All";
        if (subs_saved == 1 && rest.size() == 0) { // exactly one playlist was saved
            saved = true;
            if (top_down_naming) {
//                 cout << "\ntop-down-rename: '" << last_saved_name << "'->'" << allName << "'\n";
                rename_last_saved(allName);
//                 renamePlaylists(last_saved_name, allName);
//                 last_saved_name = allName;
            }
            else {
                ; // skip. use bottom-up name
            }
        }
        else {
            if (savePlaylist(allName, all, force_sorted, force_shuffled, min_playlist_size)) {
                saved           = true;
//                 last_saved_name = allName;
            }
        }
    }

    return saved;
}

// ---------------------------------------------------------
//      string SongContainer::get_recursive_name() const
// ---------------------------------------------------------
string SongContainer::get_recursive_name() const
{
    string upname      = "";
    if (father) upname = father->get_recursive_name();

    if (upname.length())    upname = upname+" - "+get_name();
    else                    upname = get_name();

//     cout << "get_recursive_name()='" << upname << "'\n" << flush;
    return upname;
}

// -------------------------------------------------
//      void SongContainer::warn_broken_links()
// -------------------------------------------------
void SongContainer::warn_broken_links()
{
    for (SongContainerList::iterator sc = son_begin(); sc != son_end(); ++sc) {
        (*sc)->warn_broken_links();
    }
    for (SongList::iterator so = songs.begin(); so != songs.end(); ++so) {
        SongPtr s = *so;

        if (s->isBroken()) {
            cerr << s->get_filename_maybe_link() << " is broken!\n";
        }
    }
}

// -end- of implementation of class SongContainer.

// start of implementation of class SongDir:

bool SongDir::ignore_links;
bool SongDir::read_info;
bool SongDir::scan_for_id3v2;

// -------------------------------------------------------------------------------
//      SongDir::SongDir(const string& dirname_, const SongContainer *father_)
// -------------------------------------------------------------------------------
SongDir::SongDir(const string& dirname_, const SongContainer *father_)
    : SongContainer(father_)
    , dirname(dirname_)
{
    strlist mp3s;
    strlist subdirs;

    if (show_debug_info) cout << "Entering directory '" << dirname << "'\n";

    set_name("not_detected");
    size_t slash2 = dirname.find_last_of('/');
    size_t slash1 = string::npos;
    if (slash2 != string::npos) {
        slash1 = dirname.find_last_of('/', slash2-1);
        if (slash1 != string::npos) set_name(dirname.substr(slash1+1, slash2-slash1-1));
    }

    show_dirs_read = true;

    if (ignore_links) {
        getDirContent(dirname, *pred_songs, mp3s, *pred_subdir, subdirs);
    }
    else {
        getDirContent(dirname, *pred_songs_and_links, mp3s, *pred_subdir, subdirs);
    }

    if (force_sort) set_ForceSorted(true);
    if (force_shuffle) set_ForceShuffled(true);

    for (strlist::const_iterator m = mp3s.begin(); m != mp3s.end(); ++m) {
        add_song(new Song(*m, read_info, scan_for_id3v2));
    }
    for (strlist::const_iterator s = subdirs.begin(); s != subdirs.end(); ++s) {
        add_son(new SongDir(correctDir(*s), this));
    }
    if (show_debug_info) cout << "Leaving directory '" << dirname << "'\n";
}

// -end- of implementation of class SongDir.

// start of implementation of class AllSongs:

// --------------------------------------------------------
//      void AllSongs::add_directory(const string& dir)
// --------------------------------------------------------
void AllSongs::add_directory(const string& dir)
{
    string cdir = correctDir(dir);
    directories.push_back(cdir);

    SongContainerPtr content = new SongDir(cdir, 0);
    merge(content);

    set_name("");
}


// -end- of implementation of class AllSongs.

