//  ======================================================================== //
//                                                                           //
//    File      : save.cpp                                                   //
//    Purpose   :                                                            //
//    Time-stamp: <Sun Feb/02/2003 10:13 MET Coder@ReallySoft.de>            //
//                                                                           //
//    (C) April 2002 by Ralf Westram                                         //
//                                                                           //
//    Permission to use, copy, modify, distribute and sell this software     //
//    and its documentation for any purpose is hereby granted without fee,   //
//    provided that the above copyright notice appear in all copies and      //
//    that both that copyright notice and this permission notice appear      //
//    in supporting documentation.                                           //
//                                                                           //
//    Ralf Westram makes no representations about the suitability of this    //
//    software for any purpose.  It is provided "as is" without express or   //
//    implied warranty.                                                      //
//                                                                           //
//  ======================================================================== //

#include <vector>
#include <fstream>
#include <functional>
#include <algorithm>

#include <Tools.h>
#include <Option.h>
#include <oem2ansi.h>

#include "save.h"

using namespace std;
using namespace rs;
using namespace rs::err;
using namespace rs::str;
using namespace rs::posix;

namespace playlist {

    struct savePlaylistSettings saveSettings;
     string                     last_saved_name = "";

    // -----------------------------------------------------------------------------------------------------------------------
    //      static string replace_illegal_characters(const string& s, const string& ill_chars, const string& replacement)
    // -----------------------------------------------------------------------------------------------------------------------
    static string replace_illegal_characters(const string& s, const string& ill_chars, const string& replacement) {
        string neu;

        assert(ill_chars.length() == replacement.length());

        for (string::const_iterator c = s.begin(); c != s.end(); ++c) {
            if ((*c < 32) && (*c >= 0)) {
                neu += '_';     // replace all characters below ASCII(32) by '_'
            }
            else {
                size_t pos = ill_chars.find(*c, 0);

                if (pos==string::npos) neu += *c;
                else neu                   += replacement[pos];
            }
        }

        return neu;
    }

    // ----------------------------------------------------------------------
    //      inline string cutExtension(const string& s, const char *ext)
    // ----------------------------------------------------------------------
    inline string cutExtension(const string& s, const char *ext) {
        int len = strlen(ext);

        if (stricmp(ext, s.substr(s.length()-len)) == 0) {
            return s.substr(0, s.length()-len);
        }
        return s;
    }

    // --------------------------------------------------------------------------------
    //     static string get_song_nameonly(const Song& song)
    // --------------------------------------------------------------------------------
    static string get_song_nameonly(const Song& song) {
        const string&  filename   = song.get_filename_maybe_link();
        const char    *last_slash = strrchr(filename.c_str(), '/');
        const char    *only_name  = last_slash ? last_slash+1 : filename.c_str();

        return cutExtension(cutExtension(string(only_name), ".lnk"), ".mp3");
    }

    // --------------------------------------------------------------------------------
    //     static string make_relative(const string& song, const string& relative_to_dir)
    // --------------------------------------------------------------------------------
    static string make_relative(const string& song, const string& relative_to_dir) {
        int eq = 0;

        while (song[eq]==relative_to_dir[eq]) {
            assert(song[eq]!=0); // both string exactly the same -> bad input!
            eq++;
        }

        if (eq) {
            assert(eq>=3); // X:\ should be the same
            while (song[eq-1]!='/') { // search backward for last slash
                eq--;
                assert(eq>0);
            }

            int slashes = 0;
            {
                int pos = eq;
                while (relative_to_dir[pos]) {
                    if (relative_to_dir[pos]=='/') slashes++;
                    pos++;
                }
            }

            if (!slashes) { // songs are below relative_to_dir
                return string(song, eq, string::npos);
            }

            string up;
            while (slashes--) up += "../";

            return up+string(song, eq, string::npos);
        }

        return song;
    }

    static int expand_counter = 0;
    // --------------------------------------------------------------------------------
    //     inline void expand(string& s, const string& to_replace, const string& replace_with)
    // --------------------------------------------------------------------------------
    inline void expand(string& s, const string& to_replace, const string& replace_with) {
        while (1) {
            unsigned pos = s.find(to_replace);
            if (pos==string::npos) break;
            if (pos>0) {
                s = s.substr(0, pos) + replace_with + s.substr(pos+to_replace.length());
            }
            else {
                s = replace_with+s.substr(to_replace.length());
            }
            if (replace_with.size()) ++expand_counter;
        }
    }

    // ------------------------------------------------------------------------------------------------------------------------
    //      void expand_all(string& s, const string& fullname, const string& file_or_linkname, const MPEG_Info *mpeg_info)
    // ------------------------------------------------------------------------------------------------------------------------
    void expand_all(string& s, const string& fullname, const string& file_or_linkname, const MPEG_Info *mpeg_info) {
        string Open("$(");
        string Close(")$");

        while (1) {
            string::iterator sub_entry_start = search(s.begin(), s.end(), Open.begin(), Open.end());
            if (sub_entry_start==s.end()) break;

            string::iterator after_start = sub_entry_start+2;
            string::iterator next_start  = search(after_start, s.end(), Open.begin(), Open.end());
            string::iterator next_end    = search(after_start, s.end(), Close.begin(), Close.end());

            if (next_end==s.end()) throw Error(string("Unbalanced $(..)$ in '")+s+"'");

            int opened = 1;
            while (opened) {
                assert(next_start!=next_end);

                if (next_end<next_start) {
                    --opened;
                    after_start = next_end+2;
                    next_end    = search(after_start, s.end(), Close.begin(), Close.end());
                }
                else {
                    ++opened;
                    after_start = next_start+2;
                    next_start  = search(after_start, s.end(), Open.begin(), Open.end());
                }
            }

            string expression(sub_entry_start, after_start);

            if (next_end<next_start) throw Error(string("Unexpected )$ in '")+s+"'");

            int old_expand_counter = expand_counter;
            string replace_expression(expression.begin()+2, expression.end()-2);

            expand_all(replace_expression, fullname, file_or_linkname, mpeg_info);
            int expanded = expand_counter-old_expand_counter;
            assert(expanded>=0);

            if (!expanded) replace_expression = "";
            expand(s, expression, replace_expression);
        }

        if (mpeg_info) {
            expand(s, "$s", mpeg_info->get_songname());
            expand(s, "$a", mpeg_info->get_artist());
            expand(s, "$b", mpeg_info->get_album());
            expand(s, "$y", mpeg_info->get_year());
            expand(s, "$t", mpeg_info->get_track());
            expand(s, "$q", strf("%i", mpeg_info->get_bitrate()));
        }
        else {
            const char *macros = "sabytq";
            for (int i = 0; macros[i]; ++i) {
                char macro[] = "$x";
                macro[1]   = macros[i];
                expand(s, macro, "");
            }
        }
        expand(s, "$p", fullname);
        expand(s, "$f", file_or_linkname);
    }

    static void build_playlist_entries(const Song&   song, string& normal_result, string& extended_result) {
        string song_or_linkname = get_song_nameonly(song);
        string full_song_name   = song.get_filename();

        if (saveSettings.relativePaths) full_song_name = make_relative(full_song_name, saveSettings.playlistDir);

        const MPEG_Info *mpeg_info = song.has_MPEG_info() ? &song.get_MPEG_Info() : 0;
        if (mpeg_info && !mpeg_info->has_id3()) mpeg_info = 0;

        normal_result   = saveSettings.normal_entry_pattern;
        extended_result = saveSettings.extended_entry_pattern;

        expand_all(normal_result, full_song_name, song_or_linkname, mpeg_info);
        expand_all(extended_result, full_song_name, song_or_linkname, mpeg_info);
    }

    // -----------------------------
    //      struct sortByTarget
    // -----------------------------
    struct sortByTarget {
        bool operator()(const SongPtr& s1, const SongPtr& s2) const { return s1->get_filename() < s2->get_filename(); }
    };

    //  ----------------------------------------------------------------------
    //      inline bool operator<(const SongPtr& sp1, const SongPtr& sp2)
    //  ----------------------------------------------------------------------
    inline bool operator<(const SongPtr& sp1, const SongPtr& sp2) {
        return *sp1 < *sp2;
    }

    // --------------------------------------------------------------------------------
    //     inline string outputPath(string path)
    // --------------------------------------------------------------------------------
    inline string outputPath(string path) {
        for (unsigned i=0; i<path.length(); ++i) {
            if (path[i]==BAD_DIR_SEPERATOR) {
                path[i] = GOOD_DIR_SEPERATOR;
            }
        }
        return path;
    }

    // -----------------------------------------------------
    //      inline size_t parseSize(const string& name)
    // -----------------------------------------------------
    inline size_t parseSize(const string& name) {
        size_t size = 0;
        if (saveSettings.showsize) {
            size_t close = name.find_last_of(']');
            hassert(close != string::npos);
            size_t open  = name.find_last_of('[', close-1);
            hassert(open != string::npos);

            size = atoi(name.substr(open+1, close-open-1).c_str());
        }
        return size;
    }

    // ------------------------------------------------------------------------
    //      inline string getPlaylistName(const string& name, size_t size)
    // ------------------------------------------------------------------------
    inline string getPlaylistName(const string& name, size_t size) {
        string playlist_name = name;
        if (saveSettings.showsize) {
            char buffer[10];
            sprintf(buffer, "%u", size);
            playlist_name = playlist_name + " [" + buffer + "]";
        }

        playlist_name = replace_illegal_characters(playlist_name,
                                                   "<>%?*/:|\"\\",
                                                   "()__--;-'-");

        size_t extension = playlist_name.find(PLAYLIST_EXTENSION);
        if (extension != string::npos) playlist_name.erase(extension);

        size_t last_dot = playlist_name.find_last_not_of('.');
        if (last_dot != string::npos) playlist_name.erase(last_dot+1); // erase '.''s at end of name

        playlist_name = saveSettings.playlistDir + playlist_name + PLAYLIST_EXTENSION;

        return playlist_name;
    }

    // -------------------------------------------------------
    //      void rename_last_saved(const string& newName)
    // -------------------------------------------------------
    void rename_last_saved(const string& newName) {
        size_t size         = parseSize(last_saved_name);
        string new_playlist = getPlaylistName(newName, size);

//         cout << "\ntop-level-rename: '" << last_saved_name << "'->'" << new_playlist << "'\n";

        if (rename(last_saved_name.c_str(), new_playlist.c_str()) != 0) {
            throw IOError(last_saved_name, strf("rename to '%s'", new_playlist.c_str()));
        }

        last_saved_name = new_playlist;
    }

    // -----------------------------------------------------------------------------------------------------------------------------
    //      bool savePlaylist(const string& name, const SongList& files, bool want_sorted, bool want_shuffled, size_t min_size)
    // -----------------------------------------------------------------------------------------------------------------------------
    bool savePlaylist(const string& name, const SongList& files, bool want_sorted, bool want_shuffled, size_t min_size) {

        if (files.empty()) return false; // don't save empty playlists

        // -------------------
        //      sort songs
        // -------------------
        vector<SongPtr> sorted_files;
        sorted_files.reserve(files.size());
        for (SongList::const_iterator f = files.begin(); f != files.end(); ++f) {
            SongPtr song = *f;
            if (!song->isBroken()) {
                sorted_files.push_back(song);
            }
        }

        if (sorted_files.size() < min_size) return false; // don't save playlists which are too small

        // ---------------------------------
        //      delete duplicate entries
        // ---------------------------------
        if (!saveSettings.ignoreWindowsLinks) {
            sort(sorted_files.begin(), sorted_files.end(), sortByTarget());

            vector<SongPtr> no_dups;
            size_t          size = sorted_files.size();
            no_dups.reserve(size);

            size_t i1 = 0, i2 = 1;
            no_dups.push_back(sorted_files[0]);
            while (i2<size) {
                if (sorted_files[i1]->get_filename() != sorted_files[i2]->get_filename()) no_dups.push_back(sorted_files[i2]);
                i1 = i2;
                ++i2;
            }

            sorted_files = no_dups;
        }

        if (sorted_files.size() < min_size) return false; // don't save playlists which are too small

        // -------------------------------------------
        //      sort or shuffle according to flags
        // -------------------------------------------
        if (want_sorted || (!want_shuffled && saveSettings.generally_sort)) {
            sort(sorted_files.begin(), sorted_files.end());
        }
        else if (want_shuffled || (!want_sorted && saveSettings.generally_shuffle)) {
            random_shuffle(sorted_files.begin(), sorted_files.end());
        }

        // ----------------------------
        //      build playlist name
        // ----------------------------
//         string playlist_name = name;
//         if (saveSettings.showsize) {
//             char buffer[10];
//             sprintf(buffer, "%u", sorted_files.size());
//             playlist_name = playlist_name + " [" + buffer + "]";
//         }

//         playlist_name = replace_illegal_characters(playlist_name,
//                                                    "<>%?*/:|\"\\",
//                                                    "()__--;-'-");

//         size_t extension = playlist_name.find(PLAYLIST_EXTENSION);
//         if (extension != string::npos) playlist_name.erase(extension);

//         size_t last_dot = playlist_name.find_last_not_of('.');
//         if (last_dot != string::npos) playlist_name.erase(last_dot+1); // erase '.''s at end of name

//         playlist_name = saveSettings.playlistDir + playlist_name + PLAYLIST_EXTENSION;

        string playlist_name = getPlaylistName(name, sorted_files.size());

        if (!saveSettings.playlistDirCreated) {
            create_directory(saveSettings.playlistDir);
            saveSettings.playlistDirCreated = true;
        }

        // --------------------------
        //      save the playlist
        // --------------------------
        cout << '.' << flush;
        time_t newestFile = 0;
        {
            ofstream Playlist(playlist_name.c_str());
            if (Playlist.fail()) throw IOError(playlist_name, "ofstream");
            bool     extended   = saveSettings.extended_winamp_format;

            if (extended) Playlist << "#EXTM3U" << "\n";

            vector<string> normal_info;
            vector<string> extended_info;

            if (extended) extended_info.reserve(sorted_files.size());
            normal_info.reserve(sorted_files.size());

            for (vector<SongPtr>::const_iterator ip=sorted_files.begin(); ip != sorted_files.end(); ++ip) {
                const Song& i = **ip;

                string nor_info, ext_info;

                build_playlist_entries(i, nor_info, ext_info);

                extended_info.push_back(ext_info);
                normal_info.push_back(nor_info);
            }

            int idx = 0;
            for (vector<SongPtr>::const_iterator ip=sorted_files.begin(); ip != sorted_files.end(); ++ip, ++idx) {
                const Song& i = **ip;
                if (extended) {
                    if (!i.has_MPEG_info()) {
                        Playlist << "#EXTINF:0,*** couldn't read MPEG-Info for "
                                 << get_song_nameonly(i)
                                 << "\n";
                    }
                    else {
                        Playlist << "#EXTINF:" << i.get_MPEG_Info().get_duration()-1 << ','
                                 << extended_info[idx]
                                 << "\n";
                    }
                }

                Playlist << outputPath(normal_info[idx]) << "\n";
                newestFile = max(newestFile, i.get_filedate());
            }
        }

        // set filedate of playlist to date of newest contained song
        rs::posix::setFileModTime(playlist_name, newestFile);

        last_saved_name = playlist_name; // store name for top-down renaming

        return true;
    }
};



